from torch import nn
from torch.nn import functional as F


class ConvolutionalLayer(nn.Module):
    def __init__(self, *, in_channels, out_channels,
            stride=2, kernel_size=3, padding=1, bias=False):
        super().__init__()
        self.bn = nn.BatchNorm2d(num_features=in_channels)
        self.conv = nn.Conv2d(in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                kernel_size=kernel_size,
                padding=padding,
                bias=bias)
        self.relu = nn.ReLU()

    def forward(self, x_in):
        x = self.bn(x_in)
        x = self.conv(x)
        x = self.relu(x)
        return x


class DeconvolutionalLayer(nn.Module):
    def __init__(self, *, in_channels, out_channels,
            stride=2, kernel_size=3, padding=1,
            output_padding=1, bias=False, activation=nn.ReLU()):
        super().__init__()
        self.bn = nn.BatchNorm2d(num_features=in_channels)
        self.deconv = nn.ConvTranspose2d(in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                kernel_size=kernel_size,
                padding=padding,
                output_padding=output_padding,
                bias=bias)
        self.activation = activation

    def forward(self, x_in):
        x = self.bn(x_in)
        x = self.deconv(x)
        x = self.activation(x)
        return x


class SimpleConvolutionalAutoencoder(nn.Module):
    def __init__(self, image_channels, conv_channels, conv_bias=False):
        super().__init__()
        self.conv1 = ConvolutionalLayer(
                in_channels=image_channels,
                out_channels=conv_channels[0],
                bias=conv_bias)

        self.conv2 = ConvolutionalLayer(
                in_channels=conv_channels[0],
                out_channels=conv_channels[1],
                bias=conv_bias)

        self.conv3 = ConvolutionalLayer(
                in_channels=conv_channels[1],
                out_channels=conv_channels[2],
                padding=2,
                bias=conv_bias)

        self.deconv1 = DeconvolutionalLayer(
                in_channels=conv_channels[2],
                out_channels=conv_channels[1],
                padding=2,
                output_padding=0,
                bias=conv_bias)

        self.deconv2 = DeconvolutionalLayer(
                in_channels=conv_channels[1],
                out_channels=conv_channels[0],
                bias=conv_bias)

        self.deconv3 = DeconvolutionalLayer(
                activation=nn.Sigmoid(),
                in_channels=conv_channels[0],
                out_channels=image_channels,
                bias=conv_bias)

    def encode(self, inputs, **kwargs):
        x = self.conv1(inputs)
        x = self.conv2(x)
        x_latent = self.conv3(x)
        return x_latent

    def decode(self, x_latent):
        x = self.deconv1(x_latent)
        x = self.deconv2(x)
        outputs = self.deconv3(x)
        return outputs

    def forward(self, inputs, **kwargs):
        x_latent = self.encode(inputs)
        outputs = self.decode(x_latent)
        return {'outputs': outputs}

    def calculate_loss(self, inputs, outputs, **kwargs):
        reconstruction_loss = F.binary_cross_entropy(outputs,
                inputs.view_as(outputs))
        return {'loss': reconstruction_loss}



from abc import ABC, abstractmethod


class Monitor(ABC):
    def __init__(self, name):
        self.name = name

    def start_session(self, num_steps, **kwargs):
        pass

    def end_session(self):
        pass

    @abstractmethod
    def write(self, data, step):
        pass

    def __add__(self, other):
        return CompositeMonitor(monitors=[self, other])


class CompositeMonitor(Monitor):
    def __init__(self, monitors, name='composite_monitor'):
        super().__init__(name=name)

        self.monitors = []

        for monitor in monitors:
            self.add_monitor(monitor)

    def add_monitor(self, other):
        if isinstance(other, CompositeMonitor):
            for nested_monitor in other.monitors:
                self._add_monitor(nested_monitor)
        else:
            self._add_monitor(other)

    def _add_monitor(self, other):
        if hasattr(self, other.name):
            raise RuntimeError("A monitor with the name "
                    f"'{other.name}' cannot be added to this "
                    "CompositeMonitor.  Have you already added it?")
        else:
            self.monitors.append(other)
            setattr(self, other.name, other)

    def start_session(self, num_steps, **kwargs):
        for monitor in self.monitors:
            monitor.start_session(num_steps, **kwargs)

    def end_session(self):
        for monitor in self.monitors:
            monitor.end_session()

    def write(self, data, step):
        for monitor in self.monitors:
            monitor.write(data, step)



from torch.autograd import Variable
from torch.utils import data
from torchvision import datasets, transforms

import os
import tempfile
import numpy as np


class DataGenerator:
    """
    A thin wrapper around a <data_loader> that turns torch.Tensors into
    torch.Variables (that live on cpu or on cuda) and iterates endlessly.
    """

    def __init__(self, data_loader):
        self.data_loader = data_loader
        self._inputs_cuda = False
        self._targets_cuda = False
        self.reset()

    def cuda(self, device=None):
        if isinstance(device, dict):
            for key, value in device.items():
                if key == 'inputs':
                    self._inputs_cuda = value
                elif key == 'targets':
                    self._targets_cuda = value
                else:
                    raise RuntimeError(f"Invalid key for 'device' argument: "
                            f"({key}) expected to be in ('inputs', 'targets').")
        else:
            self._inputs_cuda = device
            self._targets_cuda = device

    def reset(self):
        self._iterator = iter(self.data_loader)

    @property
    def dataset(self):
        return self.data_loader.dataset

    @property
    def batch_size(self):
        return self.data_loader.batch_size

    @property
    def num_samples(self):
        return len(self.dataset)

    def __len__(self):
        return len(self.data_loader)

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def preprocess(self, inputs, targets):
        inputs = Variable(inputs)
        targets = Variable(targets)

        if self._inputs_cuda is False:
            inputs = inputs.cpu()
        else:
            inputs = inputs.cuda(self._inputs_cuda)

        if self._targets_cuda is False:
            targets = targets.cpu()
        else:
            targets = targets.cuda(self._targets_cuda)

        return inputs, targets

    def next(self):
        try:
            inputs, targets = self._iterator.next()
        except StopIteration:
            self.reset()
            inputs, targets = self._iterator.next()
        return self.preprocess(inputs, targets)

    @classmethod
    def from_pytorch(cls,
            batch_size=500,
            data_location=None,
            num_workers=1,
            pin_memory=True,
            shuffle=True,
            train=True,
            fraction=1.0,
            transform=transforms.ToTensor(),
            dataset_class=datasets.MNIST):
        assert(fraction > 0.0)
        assert(fraction <= 1.0)

        if data_location is None:
            data_location = os.path.join(tempfile.gettempdir(),
                    dataset_class.__name__)

        dataset = dataset_class(data_location,
                train=train,
                download=True,
                transform=transform)
        indices = [int(x) for x in np.linspace(
                start=0,
                stop=len(dataset) - 1,
                num=int(len(dataset) * fraction))]
        subset = data.dataset.Subset(dataset, indices=indices)
        return cls.from_dataset(subset,
                batch_size=batch_size,
                num_workers=num_workers,
                pin_memory=pin_memory,
                shuffle=shuffle)

    @classmethod
    def from_dataset(cls, dataset, **kwargs):
        data_loader = data.DataLoader(dataset, **kwargs)
        return cls(data_loader)



from abc import ABC
from scipy import sparse
from sconce.trainer import Trainer

import seaborn as sn
import numpy as np


__all__ = ['ClassifierMixin', 'ClassifierTrainer']


class ClassifierMixin(ABC):
    def get_confusion_matrix(self, data_generator=None):
        if data_generator is None:
            data_generator = self.test_data_generator

        matrix = None
        data_generator.reset()
        for i in range(len(data_generator)):
            inputs, targets = data_generator.next()
            out_dict = self._run_model(inputs, targets, train=False)
            y_pred = np.argmax(out_dict['outputs'].cpu().data.numpy(), axis=1)
            y_true = out_dict['targets'].cpu().data.numpy()

            this_matrix = sparse.coo_matrix(
                    (np.ones(len(targets)), (y_pred, y_true)),
                    dtype='uint32').toarray()

            if matrix is None:
                matrix = this_matrix
            else:
                this_matrix, matrix = self._make_same_shape(this_matrix, matrix)
                matrix += this_matrix
        return matrix

    @staticmethod
    def _make_same_shape(a, b):
        rows = max(a.shape[0], b.shape[0])
        cols = max(a.shape[1], b.shape[1])
        padded_a = np.pad(a, ((0, rows - a.shape[0]), (0, cols - a.shape[1])),
            mode='constant', constant_values=0)
        padded_b = np.pad(b, ((0, rows - b.shape[0]), (0, cols - b.shape[1])),
            mode='constant', constant_values=0)
        return padded_a, padded_b

    def get_classification_accuracy(self, data_generator=None):
        if data_generator is None:
            data_generator = self.test_data_generator

        matrix = self.get_confusion_matrix(data_generator=data_generator)
        num_correct = np.trace(matrix)
        return num_correct / data_generator.num_samples

    def plot_confusion_matrix(self, data_generator=None, **heatmap_kwargs):
        matrix = self.get_confusion_matrix(data_generator=data_generator)

        defaults = {'cmap': 'YlGnBu', 'annot': True, 'fmt': 'd'}
        ax = sn.heatmap(matrix, **{**defaults, **heatmap_kwargs})

        ax.xaxis.set_ticklabels(ax.xaxis.get_ticklabels(), rotation=0)
        ax.yaxis.set_ticklabels(ax.yaxis.get_ticklabels(),
                rotation=0, ha='right')
        ax.set_xlabel('True')
        ax.set_ylabel('Predicted')
        return ax

    def plot_misclassified_samples(self, true_label):
        pass


class ClassifierTrainer(Trainer, ClassifierMixin):
    pass
